﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using ClientServer.Controllers.Core.Login;
using ClientServer.Db;
using ClientServer.Helpers;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Hosting;
using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.AspNetCore.Session;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Http.Features;
using Microsoft.Extensions.Caching.Distributed;
using Newtonsoft.Json.Serialization;

namespace ClientServer
{
    public class Startup
    {
        public Startup(IHostingEnvironment env)
        {
            var builder = new ConfigurationBuilder()
                .SetBasePath(env.ContentRootPath)
                .AddJsonFile("appsettings.json", optional: false,
                    reloadOnChange: false) //reloadOnChange not working properly... https://andrewlock.net/reloading-strongly-typed-options-in-asp-net-core-1-1-0/
                //.AddJsonFile($"appsettings.{env.EnvironmentName}.json", optional: true)
                .AddEnvironmentVariables();

            ConfigurationRoot = builder.Build();


            //apply appsettings.json
            AppConfiguration.SetFromAppSettings(ConfigurationRoot);


            //we need this for migration or if we copied dirs and have different number of users (old users that have no upload dir)...
            //if we have manyyyyyy users this slowdowns startup hard!!
            if (AppConfiguration.EnsureFileUploadDirStructure)
            {
                try
                {
                    Files.CreateUploadFilesStructureIfNotExists();
                }
                catch (Exception e)
                {
                    Console.WriteLine("[ERROR] Cannot create upload file structure");
                    Console.WriteLine(e);
                    throw;
                }
            }
            else
            {
                Console.WriteLine("[INFO] skipping upload file dir structure check");
            }

            //DISABLE THIS BEFORE MAKEING (AND APPLYING) A NEW MIGRATION 

            if (AppConfiguration.IsDbBootstrapEnabled)
            {
                Console.WriteLine("YAPEX: Bootstrap is enabled!");
                BootstrapDb();
            }
            else
            {
                Console.WriteLine("YAPEX: Bootstrap is disabled");
            }
        }

        /// <summary>
        /// ensures that the database is create and migrated
        /// </summary>
        private void BootstrapDb()
        {
            #region ensure that the database is up to date...

            var optionsBuilder = new DbContextOptionsBuilder<YapexDbContext>();
            optionsBuilder.UseNpgsql(AppConfiguration.DbConnectionString);
            try
            {
                using (var context = new YapexDbContext(optionsBuilder.Options))
                {
                    context.Database.Migrate();
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("error migrating database: " + ex.Message.ToString());
                if (ex.InnerException != null)
                {
                    Console.WriteLine("inner exception: " + ex.InnerException.Message.ToString());
                }

                Environment.Exit(1);
            }

            #endregion
        }

        public IConfigurationRoot ConfigurationRoot { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services) //IServiceProvider
        {
            services.AddCors(options =>
            {
                options.AddPolicy("CorsPolicy",
                    builder => builder
                        .WithOrigins(AppConfiguration.DebugOriginUrl)
                        .WithOrigins(AppConfiguration.DeployOriginUrls.ToArray())
                        .AllowAnyMethod()
                        .AllowAnyHeader()
                        .AllowCredentials());
            });

            services
                .AddDbContext<YapexDbContext>(option => option.UseNpgsql(AppConfiguration.DbConnectionString),
                    ServiceLifetime.Scoped);

            //services.AddDbContext<SyndromDbContext>(options =>
            //{
            //    options.UseSqlite(DbConnectionStringSqlite);
            //});


            services.Configure<FormOptions>(o =>
            {
                o.MultipartBodyLengthLimit = AppConfiguration.MaxUploadFileSizeInByte;
            });

            // Add framework services.
            services.AddMvc(options =>
            {
                /* TODO does not work...
                var formatter = options.OutputFormatters.First(p => p is JsonOutputFormatter) as JsonOutputFormatter;
                formatter.SerializerSettings.ContractResolver = new CamelCasePropertyNamesContractResolver();
                formatter.SerializerSettings.DefaultValueHandling = Newtonsoft.Json.DefaultValueHandling.Ignore;
                */
            });


            //services.AddDistributedMemoryCache();
            //services.AddSingleton<IDistributedCache, RedisCache>();

            //http://benjii.me/2016/07/using-sessions-and-httpcontext-in-aspnetcore-and-mvc-core/
            //    "Microsoft.Framework.Caching.Redis":  "1.0.0-*"
            //https://github.com/aspnet/Caching/issues/121
            //services.AddDistributedRedisCache(options =>
            //{
            //    options.Configuration = "localhost";
            //    options.InstanceName = "SampleInstance";
            //});


            //services.AddSession(options =>
            //{
            //    options.CookieName = "id";
            //    options.CookieHttpOnly = true;
            //    options.IdleTimeout = TimeSpan.FromMinutes(30); //
            //});


            //var builder = new ContainerBuilder();

            //builder.Populate(services);
            //var container = builder.Build();


            //return container.Resolve<IServiceProvider>();
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IHostingEnvironment env, ILoggerFactory loggerFactory)
        {
            //loggerFactory.AddConsole(Configuration.GetSection("Logging"));
            //loggerFactory.AddDebug();

            //TODO maybe disable this when we use nginx ??
            //must be before use static files else index_deploy would not get this header
            app.Use(async (context, next) =>
            {
                context.Response.Headers.Add("X-Frame-Options", "DENY");
                await next();
            });

            //TODO maybe disable this when we use nginx ??
            app.UseStaticFiles();

            app.UseCors("CorsPolicy");

            //Microsoft.Framework.Caching.Redis
            // Redis implementation of IDistributedCache.
            // This will override any previously registered IDistributedCache service.
            //services.AddSingleton<IDistributedCache, RedisCache>();

            //before mvc else the session wont be injected in the controllers!!
            //app.UseSession();

            app.Use(async (context, next) =>
            {
                if (context.Request.Path.HasValue
                    && context.Request.Path.Value.StartsWith("/" + Constants.InitControllerAbsolutePath)
                    && AppConfiguration.IsInitControllerEnabled == false)
                {
                    context.Response.StatusCode = 409;
                    await context.Response.WriteAsync("init is disabled!");
                    return;
                }

                // Do work that doesn't write to the Response.
                if (context.Request.Path.HasValue && context.Request.Path.Value.StartsWith("/" + Constants.ApiPrefix))
                {
                    //for firefox we need to set return type to json explicitly else we get an error?
                    //but actually it works most of the time
                    context.Response.ContentType = "application/json";
                }

                try
                {
                    await next.Invoke();
                }
                catch (InvalidDataException e) //only catches this exception type
                {
//                    Console.WriteLine(e);
                    context.Response.ContentType = "application/json";
                    context.Response.StatusCode = 200;
                    await
                        context.Response.WriteAsync(
                            Jc.Serialize(
                                new BasicResponse(ResponseCode.ServerError, $"file content size (multipart body) exceeded the limit of " +
                                                                            $"{AppConfiguration.MaxUploadFileSizeInByte} Bytes (~{Files.GetHumanReadableSize(AppConfiguration.MaxUploadFileSizeInByte)})")));
                    return;
                }
               
                // Do logging or other work that doesn't write to the Response.
            });

            app.UseMvc();
        }
    }
}
